/*----------------------------------------------------------------------------/
/* Vector Tracer R0.06a                                       (C)ChaN, 2005                                                       */
/-----------------------------------------------------------------------------/
/ Vector Tracer is a simple vector frame creator for laser display systems.
/ This is a free software that opened for education, research and commercial
/ developments under license policy of following terms.
/
/  Copyright (C) 2011, ChaN, all right reserved.
/
/ * The mkv2 is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/ * Redistributions in binary form must describe the above copyright notice
/   in the documentation.
/
/----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <commctrl.h>
#include "resource.h"

#define MAX_VECTORS	3000	/* Number of vectors per frame */
#define MAX_FRAMES	1000	/* Number of frames */


/* Type of entry in vector buffer */
#pragma pack(1)

typedef struct _vects {
	short	f, x, y;
} VECTS;

typedef struct {
	char	sign[7];	// "ILDA", 0, 0, 0
	BYTE	format;		// 0:3D frame, 1:2D frame, 2:Pallete
	char	framename[16];
	WORD	points;
	WORD	framenum;
	WORD	totalframes;
	BYTE	head;
	BYTE	res;
} ILDA;

#define LEWORD(a)	((WORD)(((a) >> 8) & 0xFF)|(((a) << 8) & 0xFF00))	// endian conversion

#pragma pack()



/* Strings */
const char *ProgName = "xN";
const char *VerStr = "Vector Tracer R0.06a\n(C)ChaN, 2006";


/* File Control */
char FileName[_MAX_PATH], DispName[_MAX_FNAME];
BOOL fFileChanged, fFrameChanged;
DWORD FilterIdx = 1;
const char *sFileExt[] = { "", "csv", "ild", "frm", "" };


/* Background image, Scroll controls */
fImgLoaded = TRUE;
HBITMAP hImage;
POINT ImgSize, DispOfs;
int nZoom = 6;


/* System handles */
HWND hMainWnd, hSBWnd, hDrawWnd, hVectWnd, hVectViewWnd, hFrameWnd, hFrameViewWnd;
HMENU hMenu;
HACCEL hAccel;
HINSTANCE hMainInst;


/* Dialog data buffers */
char *DlgCaption[2];
int DlgMode;
int DlgValue[3];
char DlgStr1[20];


/* Vector buffers and related variables */
int nPps = 30000, nDelay = 3;
int CurrentFrame, nFrames;
int nVectors, nVectorPrev, nVectorSaved;
int VectorsLst[MAX_FRAMES];
char StrLst[MAX_FRAMES][16];
VECTS FrameLst[MAX_FRAMES][MAX_VECTORS];
VECTS VectLst[MAX_VECTORS], VectLstPrev[MAX_VECTORS];






LRESULT CALLBACK
OpenFileProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	char fname[_MAX_FNAME], *pfn;
	static LPOPENFILENAME lpOpenFname;


	switch (msg) {
		case WM_INITDIALOG :
			lpOpenFname = (LPOPENFILENAME)lParam;
			return 1;

		case WM_NOTIFY :
			switch (((LPNMHDR)lParam)->code) {
				case CDN_TYPECHANGE :
					SendMessage(GetParent(hWnd), CDM_GETSPEC, (WPARAM)sizeof(fname), (LPARAM)fname);
					if (strlen(fname)) {
						pfn = strstr(fname, ".");
						if (pfn == NULL) pfn = &fname[strlen(fname)];
						sprintf(pfn, ".%s", sFileExt[lpOpenFname->nFilterIndex]);
						SendMessage(GetParent(hWnd), CDM_SETCONTROLTEXT, (WPARAM)edt1, (LPARAM)fname);
					}
					SendMessage(GetParent(hWnd), CDM_SETDEFEXT, 0, (LPARAM)sFileExt[lpOpenFname->nFilterIndex]);
				return 1;
			}
		break;
	}

	return 0;
}




LRESULT CALLBACK
OptWndProc(HWND hWnd, UINT msg,  WPARAM wParam, LPARAM lParam)
{
	switch (msg) {
		case WM_INITDIALOG :
			SendMessage(hWnd, WM_SETTEXT, 0, (LPARAM)DlgCaption[0]);
			SendMessage(GetDlgItem(hWnd, IDC_STATIC1), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)DlgCaption[1]);
			if (DlgMode) {
				ShowWindow(GetDlgItem(hWnd, IDC_EDIT2), SW_HIDE);
				SendMessage(GetDlgItem(hWnd, IDC_EDIT1), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)DlgStr1);
				SetFocus(GetDlgItem(hWnd, IDC_EDIT1));
			} else {
				ShowWindow(GetDlgItem(hWnd, IDC_EDIT1), SW_HIDE);
				SendMessage(GetDlgItem(hWnd, IDC_EDIT2), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)DlgStr1);
				SetFocus(GetDlgItem(hWnd, IDC_EDIT2));
			}
			return (0);
		case WM_COMMAND :
			switch (LOWORD(wParam)) {
				case IDOK :
					if (DlgMode)
						GetWindowText(GetDlgItem(hWnd, IDC_EDIT1), DlgStr1, sizeof(DlgStr1));
					else
						GetWindowText(GetDlgItem(hWnd, IDC_EDIT2), DlgStr1, sizeof(DlgStr1));
					EndDialog(hWnd, IDOK);
					break;
				case IDCANCEL :
					EndDialog(hWnd, IDCANCEL);
					break;
				default :
					return FALSE;
			}
		default :
			return FALSE;
	}
	return TRUE;

}




LRESULT CALLBACK
VectEditWndProc(HWND hWnd, UINT msg,  WPARAM wParam, LPARAM lParam)
{
	char str[20];


	switch (msg) {
		case WM_INITDIALOG :
			SendMessage(GetDlgItem(hWnd, IDC_STATICN), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)DlgStr1);
			_itoa(DlgValue[0], str, 10);
			SendMessage(GetDlgItem(hWnd, IDC_EDITV), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)str);
			_itoa(DlgValue[1], str, 10);
			SendMessage(GetDlgItem(hWnd, IDC_EDITX), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)str);
			_itoa(DlgValue[2], str, 10);
			SendMessage(GetDlgItem(hWnd, IDC_EDITY), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)str);
			SetFocus(GetDlgItem(hWnd, IDC_EDITV));
			return FALSE;

		case WM_COMMAND :
			switch (LOWORD(wParam)) {
				case IDOK :
					GetWindowText(GetDlgItem(hWnd, IDC_EDITV), str, sizeof(str));
					DlgValue[0] = atoi(str);
					GetWindowText(GetDlgItem(hWnd, IDC_EDITX), str, sizeof(str));
					DlgValue[1] = atoi(str);
					GetWindowText(GetDlgItem(hWnd, IDC_EDITY), str, sizeof(str));
					DlgValue[2] = atoi(str);
					EndDialog(hWnd, IDOK);
					break;
				case IDCANCEL :
					EndDialog(hWnd, IDCANCEL);
					break;
				default :
					return FALSE;
			}

		default :
			return FALSE;
	}
	return TRUE;

}




void UpdateCaption(int mode) {
	char str[_MAX_FNAME+20];


	if (mode & 1) {
		sprintf(str, "%s%s%s", DispName, (strlen(DispName) ? " - " : ""), ProgName);
		SetWindowText(hMainWnd, str);
	}
	if (mode & 2) {
		if (nVectors)
			sprintf(str, " %d points,  %-1.1ffps/%dpps", nVectors, (double)nPps / nVectors, nPps);
		else
			strcpy(str, "No point");
		SendMessage(hSBWnd, SB_SETTEXT, 0, (LPARAM)str);
		sprintf(str, "Frame: %d/%d", CurrentFrame+1, nFrames);
		SendMessage(hSBWnd, SB_SETTEXT, 1, (LPARAM)str);
		sprintf(str, "Zoom: 1/%d", 1 << nZoom);
		SendMessage(hSBWnd, SB_SETTEXT, 2, (LPARAM)str);
	}
}




void SetChanged (BOOL mode)
{
	if (mode) {
		fFrameChanged = TRUE;
		fFileChanged = TRUE;
		EnableMenuItem(GetMenu(hMainWnd), IDM_SAVE, MF_BYCOMMAND | MF_ENABLED);
		EnableMenuItem(GetMenu(hMainWnd), IDM_UNDO, MF_BYCOMMAND | ((nVectorPrev == -1) ? MF_GRAYED : MF_ENABLED));
	} else {
		fFrameChanged = FALSE;
		fFileChanged = FALSE;
		EnableMenuItem(GetMenu(hMainWnd), IDM_SAVE, MF_BYCOMMAND | MF_GRAYED);
	}
}




void CreateUndo (int mode	/* 0:Vector Added, 1:Large Changed, 2:Invalidate */
				)
{
	switch (mode) {
		case 0 :
			nVectorPrev = nVectors;
			break;
		case 1 :
			memcpy(&VectLstPrev[0], &VectLst[0], sizeof(VECTS) * nVectors);
			nVectorSaved = nVectors;
			nVectorPrev = -2;
			break;
		case 2 :
			nVectorPrev = -1;
			EnableMenuItem(GetMenu(hMainWnd), IDM_UNDO, MF_BYCOMMAND | MF_GRAYED);
	}
}




BOOL Undo ()
{
	if (nVectors == -1) return FALSE;

	if (nVectorPrev >= 0) {
		nVectors = nVectorPrev;
	} else {
		nVectors = nVectorSaved;
		memcpy(&VectLst[0], &VectLstPrev[0], sizeof(VECTS) * nVectorSaved);
	}
	nVectorPrev = -1;
	fFrameChanged = TRUE;

	return TRUE;
}




void GetFrame (int frm)
{
	int i;


	if (frm >= MAX_FRAMES) return;

	nVectors = VectorsLst[frm];
	memcpy(&VectLst[0], &FrameLst[frm][0], sizeof(VECTS) * MAX_VECTORS);
	for (i = 0; i < nVectors; i++) VectLst[i].f &= 1;

	fFrameChanged = FALSE;
}




void PutFrame (int frm, int mode)
{
	if (frm >= MAX_FRAMES) return;
	if (!mode && !fFrameChanged) return;

	while (nFrames <= frm) {
		VectorsLst[nFrames] = 0;
		memset(&StrLst[nFrames][0], ' ', 16); 
		nFrames++;
	}

	VectorsLst[frm] = nVectors;
	memcpy(&FrameLst[frm][0], &VectLst[0], sizeof(VECTS) * MAX_VECTORS);

	fFrameChanged = FALSE;
}




BOOL InsertBlankFrame (int frm)
{
	if (nFrames >= MAX_FRAMES) return FALSE;
	if (frm >= MAX_FRAMES-1) return FALSE;

	while (nFrames < frm) {
		VectorsLst[nFrames] = 0;
		memset(&StrLst[nFrames][0], ' ', 16); 
		nFrames++;
	}

	memmove(&VectorsLst[frm+1], &VectorsLst[frm], (nFrames - frm) * sizeof(int));
	memmove(&FrameLst[frm+1][0], &FrameLst[frm][0], (nFrames - frm) * sizeof(VECTS) * MAX_VECTORS);
	memmove(&StrLst[frm+1][0], &StrLst[frm][0], (nFrames - frm) * sizeof(StrLst[0]));
	VectorsLst[frm] = 0;
	memset(&StrLst[frm][0], ' ', sizeof(StrLst[0]));
	nFrames++;

	return TRUE;
}




BOOL DeleteFrame (int frm)
{
	if (frm >= nFrames) return FALSE;
	nFrames--;

	memmove(&VectorsLst[frm], &VectorsLst[frm+1], (nFrames - frm) * sizeof(int));
	memmove(&FrameLst[frm][0], &FrameLst[frm+1][0], (nFrames - frm) * sizeof(VECTS) * MAX_VECTORS);
	memmove(&StrLst[frm][0], &StrLst[frm+1][0], (nFrames - frm) * sizeof(StrLst[0]));
	VectorsLst[nFrames] = 0;

	return TRUE;
}




int GetFrameSize (int f_from, int f_to)
{
	int frm, n, max = 0;


	for (frm = f_from; frm <= f_to; frm++) {
		for (n = 0; n < VectorsLst[frm]; n++) {
			if (max < FrameLst[frm][n].x) max = FrameLst[frm][n].x;
			if (max < -FrameLst[frm][n].x) max = -FrameLst[frm][n].x;
			if (max < FrameLst[frm][n].y) max = FrameLst[frm][n].y;
			if (max < -FrameLst[frm][n].y) max = -FrameLst[frm][n].y;
		}
	}
	return max * 2;
}




BOOL ResizeFrames (int resize, int f_from, int f_to)
{
	int frm, n, max = 0;
	double mag;


	if ((resize <= 0) || (resize >= 65536)) return FALSE;

	if (resize < 100)
		mag = (double)resize / 100;
	else
		mag = (double)resize / GetFrameSize(f_from, f_to);

	for (frm = f_from; frm <= f_to; frm++) {
		for (n = 0; n < VectorsLst[frm]; n++) {
			FrameLst[frm][n].x = (short)(FrameLst[frm][n].x * mag);
			FrameLst[frm][n].y = (short)(FrameLst[frm][n].y * mag);
		}
	}

	return TRUE;
}




void LV_RfshVector (int mode	// b0: retain last item's position
					) {			// b1: without change of number of items
	int i, s = -1;
	char str[10];
	LVITEM lvitm;


	if (!hVectViewWnd) return;

	if (mode & 2) {
		for (i = 0; i < nVectors; i++) {
			_itoa(VectLst[i].x, str, 10);
			ListView_SetItemText(hVectViewWnd, i, 2, str);
			_itoa(-VectLst[i].y, str, 10);
			ListView_SetItemText(hVectViewWnd, i, 3, str);
			ListView_SetItemState(hVectViewWnd, i, (VectLst[i].f & 2) ? LVIS_SELECTED : 0, LVIS_SELECTED);

		}
		return;
	}
	
	if (mode & 1)
		s = ListView_GetTopIndex(hVectViewWnd);

	ShowWindow(hVectViewWnd, SW_HIDE);
	ListView_DeleteAllItems(hVectViewWnd);
	ListView_SetItemCount(hVectViewWnd, nVectors);
	lvitm.mask = LVIF_TEXT;
	lvitm.pszText = str;
	lvitm.iSubItem = 0;
	for (i = 0; i < nVectors; i++) {
		lvitm.iItem = i;
		_itoa(i+1, str, 10);
		ListView_InsertItem(hVectViewWnd, &lvitm);
		_itoa(VectLst[i].f & 1, str, 10);
		ListView_SetItemText(hVectViewWnd, i, 1, str);
		_itoa(VectLst[i].x, str, 10);
		ListView_SetItemText(hVectViewWnd, i, 2, str);
		_itoa(-VectLst[i].y, str, 10);
		ListView_SetItemText(hVectViewWnd, i, 3, str);
		if (VectLst[i].f & 2)
			ListView_SetItemState(hVectViewWnd, i, LVIS_SELECTED, LVIS_SELECTED);
	}
	ShowWindow(hVectViewWnd, SW_SHOW);

	if (mode & 1) {
		i = HIWORD(ListView_GetItemSpacing(hVectViewWnd, TRUE));
		ListView_Scroll(hVectViewWnd, 0, s * i);
	}
	else
		ListView_EnsureVisible(hVectViewWnd, i-1, FALSE);

}




void LV_InsertVector ()
{
	int i, n;


	if (!hVectViewWnd) return;
	if ((nVectors == 0) || (nVectors >= MAX_VECTORS)) return;
	i = ListView_GetNextItem(hVectViewWnd, -1, LVNI_FOCUSED);
	if (i == -1) return;

	CreateUndo(1);

	for (n = nVectors; n > i; n--)
		VectLst[n] = VectLst[n-1];
	nVectors++;


	SetChanged(TRUE);
	UpdateCaption(2);
	LV_RfshVector(1);
	ListView_SetItemState(hVectViewWnd, i, LVIS_FOCUSED, LVIS_FOCUSED);
	InvalidateRect(hDrawWnd, NULL, TRUE);
}




void LV_DeleteVectors ()
{
	int i, n;


	if (!hVectViewWnd) return;
	i = ListView_GetNextItem(hVectViewWnd, -1, LVNI_SELECTED);
	if (i == -1) return;

	CreateUndo(1);

	do {
		VectLst[i].f |= 4;
		i = ListView_GetNextItem(hVectViewWnd, i, LVNI_SELECTED);
	} while (i >= 0);

	for (i = nVectors-1; i >= 0; i--) {
		if (VectLst[i].f & 4) {
			nVectors--;
			for (n = i; n < nVectors; n++)
				VectLst[n] = VectLst[n+1];
		}
	}

	SetChanged(TRUE);
	UpdateCaption(2);
	LV_RfshVector(1);
	InvalidateRect(hDrawWnd, NULL, TRUE);
}




void FV_RfshFrame (int mode) {
	int i;
	char str[20];
	LVITEM fvitm;


	if (!hFrameViewWnd) return;

	if (mode == 0) {
		if (CurrentFrame < nFrames)
			ListView_SetItemState(hFrameViewWnd, CurrentFrame, LVIS_FOCUSED, LVIS_FOCUSED);
			ListView_EnsureVisible(hFrameViewWnd, CurrentFrame, FALSE);
		return;
	}

	ShowWindow(hFrameViewWnd, SW_HIDE);
	ListView_DeleteAllItems(hFrameViewWnd);
	ListView_SetItemCount(hFrameViewWnd, nFrames);
	fvitm.mask = LVIF_TEXT;
	fvitm.pszText = str;
	fvitm.iSubItem = 0;
	for (i = 0; i < nFrames; i++) {
		fvitm.iItem = i;
		_itoa(i+1, str, 10);
		ListView_InsertItem(hFrameViewWnd, &fvitm);
		str[16] = 0;
		memcpy(str, &StrLst[i][0], 16);
		ListView_SetItemText(hFrameViewWnd, i, 1, str);
	}
	ShowWindow(hFrameViewWnd, SW_SHOW);

	if (CurrentFrame < nFrames) {
		i = HIWORD(ListView_GetItemSpacing(hFrameViewWnd, TRUE));
		ListView_Scroll(hFrameViewWnd, 0, CurrentFrame * i);
		ListView_SetItemState(hFrameViewWnd, CurrentFrame, LVIS_FOCUSED, LVIS_FOCUSED);
	}
}




void FV_InsertFrame ()
{
	int i;


	if (!hFrameViewWnd) return;

	if ((nFrames == 0) || (nFrames >= MAX_FRAMES)) return;
	i = ListView_GetNextItem(hFrameViewWnd, -1, LVNI_FOCUSED);
	if (i == -1) return;

	InsertBlankFrame(i);
	memmove(&StrLst[i][0], &StrLst[i+1][0], sizeof(StrLst[0]));

	SetChanged(TRUE);
	PutFrame(i, 1);
	CreateUndo(2);
	UpdateCaption(2);
	FV_RfshFrame(1);
}




void FV_DeleteFrames ()
{
	int i;


	if (!hFrameViewWnd) return;

	i = ListView_GetNextItem(hFrameViewWnd, -1, LVNI_SELECTED);
	if (i == -1) return;

	do {
		VectorsLst[i] = -1;
		i = ListView_GetNextItem(hFrameViewWnd, i, LVNI_SELECTED);
	} while (i >= 0);

	for (i = nFrames-1; i >= 0; i--) {
		if (VectorsLst[i] == -1)
			DeleteFrame(i);
	}

	SetChanged(TRUE);
	GetFrame(CurrentFrame);
	CreateUndo(2);
	UpdateCaption(2);
	FV_RfshFrame(1);
	LV_RfshVector(1);
	InvalidateRect(hDrawWnd, NULL, TRUE);
}




BOOL FV_GetSelection(int *f_from, int *f_to)
{
	int i;


	if (!hFrameViewWnd) return FALSE;

	i = ListView_GetNextItem(hFrameViewWnd, -1, LVNI_SELECTED);
	if (i == -1) return FALSE;

	*f_from = i;
	while (i >= 0) {
		*f_to = i;
		i = ListView_GetNextItem(hFrameViewWnd, i, LVNI_SELECTED);
	}

	return TRUE;
}




BOOL FV_ResizeFrames (HWND hWnd)
{
	int f_from, f_to;


	if (!FV_GetSelection(&f_from, &f_to)) return FALSE;

	DlgCaption[0] = "Resize Frames";
	DlgCaption[1] = "Change frame size to:";
	DlgMode = 0;
	PutFrame(CurrentFrame, 0);
	_itoa(GetFrameSize(f_from, f_to), DlgStr1, 10);
	if (DialogBox(hMainInst, MAKEINTRESOURCE(IDD_OPTION), hWnd, (DLGPROC)OptWndProc)
		!= IDOK) return FALSE;

	if (!ResizeFrames(atoi(DlgStr1), f_from, f_to)) return FALSE;

	CreateUndo(2);
	SetChanged(TRUE);
	GetFrame(CurrentFrame);
	UpdateCaption(2);
	LV_RfshVector(1);
	InvalidateRect(hDrawWnd, NULL, TRUE);

	return TRUE;
}




void FV_EditFrame (HWND hWnd)
{
	int i, l;


	if (!hFrameViewWnd) return;
	i = ListView_GetNextItem(hFrameViewWnd, -1, LVNI_FOCUSED);
	if (i == -1) return;


	DlgCaption[0] = "Frame Name";
	DlgCaption[1] = "Enter frame name:";
	DlgMode = 1;
	ListView_GetItemText(hFrameViewWnd, i, 1, DlgStr1, sizeof(DlgStr1));
	if (DialogBox(hMainInst, MAKEINTRESOURCE(IDD_OPTION), hWnd, (DLGPROC)OptWndProc)
		== IDOK) {
		for (l = strlen(DlgStr1); l < 16; l++) DlgStr1[l] = ' ';
		memcpy(&StrLst[i][0], DlgStr1, sizeof(StrLst[0]));
		ListView_SetItemText(hFrameViewWnd, i, 1, DlgStr1);
		SetChanged(TRUE);
	}
}




void LV_EditVector (HWND hWnd)
{
	int i;
	char str2[20];
	UINT f;


	if (!hVectViewWnd) return;
	i = ListView_GetNextItem(hVectViewWnd, -1, LVNI_FOCUSED);
	if (i == -1) return;

	_itoa(i+1, DlgStr1, 10);
	ListView_GetItemText(hVectViewWnd, i, 1, str2, sizeof(str2));
	DlgValue[0] = atoi(str2);
	ListView_GetItemText(hVectViewWnd, i, 2, str2, sizeof(str2));
	DlgValue[1] = atoi(str2);
	ListView_GetItemText(hVectViewWnd, i, 3, str2, sizeof(str2));
	DlgValue[2] = atoi(str2);

	if (DialogBox(hMainInst, MAKEINTRESOURCE(IDD_VECTEDIT), hWnd, (DLGPROC)VectEditWndProc)
		== IDOK) {
		CreateUndo(1);

		for (i = 0; i < nVectors; i++) {
			f = ListView_GetItemState(hVectViewWnd, i, LVIS_SELECTED | LVIS_FOCUSED);
			if (f & LVIS_FOCUSED) {
				VectLst[i].f = DlgValue[0] & 1;
				_itoa(VectLst[i].f & 1, str2, 10);
				ListView_SetItemText(hVectViewWnd, i, 1, str2);
			}
			if (f & (LVIS_FOCUSED | LVIS_SELECTED)) {
				VectLst[i].x = DlgValue[1];
				_itoa(VectLst[i].x, str2, 10);
				ListView_SetItemText(hVectViewWnd, i, 2, str2);
				VectLst[i].y = -DlgValue[2];
				_itoa(-VectLst[i].y, str2, 10);
				ListView_SetItemText(hVectViewWnd, i, 3, str2);
			}
		}

		SetChanged(TRUE);
		InvalidateRect(hDrawWnd, NULL, TRUE);
	}

}




void ChangeFrame (int frm)
{
	PutFrame(CurrentFrame, 0);
	CurrentFrame = frm;
	GetFrame(CurrentFrame);
	CreateUndo(2);
	UpdateCaption(2);
	FV_RfshFrame(0);
	LV_RfshVector(1);
	InvalidateRect(hDrawWnd, NULL, TRUE);
}




BOOL AddVector (int mode, WORD x, WORD y)
{
	HDC hdc;
	HPEN hPen, hPenPrev;
	HBRUSH hBrush, hBrushPrev;
	LONG px, py, i;
	char str[16];
	LVITEM lvitm;


	if (nVectors >= MAX_VECTORS) return FALSE;
	px = (x + DispOfs.x) << nZoom;
	if ((px < -32768) || (px > 32767)) return FALSE;
	py = (y + DispOfs.y) << nZoom;
	if ((py < -32768) || (py > 32767)) return FALSE;

	CreateUndo(0);

	hdc = GetDC(hDrawWnd);
	if (mode)
		hPen = CreatePen(PS_SOLID, 1, RGB(0,255,0));
	else
		hPen = CreatePen(PS_SOLID, 1, RGB(0,0,255));
	hPenPrev = SelectObject(hdc, hPen);
	hBrush = GetStockObject(NULL_BRUSH);
	hBrushPrev = SelectObject(hdc, hBrush);

	if (nVectors) {
		px = (VectLst[nVectors-1].x >> nZoom) - DispOfs.x;
		py = (VectLst[nVectors-1].y >> nZoom) - DispOfs.y;

		/* Align steady points */
		if ((x - px <= 2) && (x - px >= -2) &&
			(y - py <= 2) && (y - py >= -2)) {
			x = (WORD)px;
			y = (WORD)py;
		}

		MoveToEx(hdc, px, py, NULL);
		LineTo(hdc, x, y);
	}
	Ellipse(hdc, x - 2, y - 2, x + 3, y + 3);

	SelectObject(hdc, hBrushPrev);
	SelectObject(hdc, hPenPrev);
	DeleteObject(hPen);
	ReleaseDC(hDrawWnd, hdc);

	VectLst[nVectors].f = mode & 1;
	VectLst[nVectors].x = (short)((x + DispOfs.x) << nZoom);
	VectLst[nVectors].y = (short)((y + DispOfs.y) << nZoom);


	if (hVectViewWnd) {
		lvitm.mask = LVIF_TEXT;
		lvitm.pszText = str;
		lvitm.iSubItem = 0;
		lvitm.iItem = nVectors;
		_itoa(nVectors+1, str, 10);
		i = ListView_InsertItem(hVectViewWnd, &lvitm);
		if (i != -1) {
			_itoa(VectLst[i].f & 1, str, 10);
			ListView_SetItemText(hVectViewWnd, i, 1, str);
			_itoa(VectLst[i].x, str, 10);
			ListView_SetItemText(hVectViewWnd, i, 2, str);
			_itoa(-VectLst[i].y, str, 10);
			ListView_SetItemText(hVectViewWnd, i, 3, str);
			ListView_EnsureVisible(hVectViewWnd, i, FALSE);
		}
	}

	nVectors++;

	return TRUE;
}




BOOL SubtractVector (WORD x, WORD y)
{
	LONG px, py, tx, ty, i;


	if (nVectors == 0) return FALSE;
	CreateUndo(0);

	px = x + DispOfs.x; py = y + DispOfs.y;
	for (i = nVectors-1; i >= 0; i--) {
		tx = px - (VectLst[i].x >> nZoom);
		ty = py - (VectLst[i].y >> nZoom);
		if ((tx <= 3) && (tx >= -3) &&
			(ty <= 3) && (ty >= -3)) break;
	}

	if (i >= 0)
		nVectors = i;
	else
		nVectors--;

	if (hVectViewWnd) {
		i = ListView_GetItemCount(hVectViewWnd) - 1;
		ShowWindow(hVectViewWnd, SW_HIDE);
		for ( ; i >= nVectors ; i--)
				ListView_DeleteItem(hVectViewWnd, i);
		ShowWindow(hVectViewWnd, SW_SHOW);
	}
	return TRUE;
}




BOOL DragDragStart(long x, long y)
{
	long sx, sy;
	int i, chg = 0;
	BOOL res = FALSE;


	x += DispOfs.x;
	y += DispOfs.y;

	for (i = 0; i < nVectors; i++) {
		sx = (VectLst[i].x >> nZoom) - x;
		sy = (VectLst[i].y >> nZoom) - y;
		VectLst[i].f &= 3;
		if ((sx <= 2) && (sx >= -2) &&
			(sy <= 2) && (sy >= -2)) {
			if (VectLst[i].f & 2) {
				chg |= 1;
			} else {
				chg |= 2; VectLst[i].f |= 4;
			}
		}
	}

	if (chg) {
		if (chg & 1) {
			for (i = 0; i < nVectors; i++) {
				if (VectLst[i].f & 2)
					VectLst[i].f |= 4;
				else
					VectLst[i].f &= ~4;
			}
		}
		res = TRUE;
	}

	return res;
}




void DragDragging (int mode, long x, long y)
{
	static long px, py;
	long vx, vy, sx = 100000, sy = 100000;
	HDC hdc;
	HPEN hPen, hPenPrev;
	HBRUSH hBrush, hBrushPrev;
	int i;


	hdc = GetDC(hDrawWnd);
	SetROP2(hdc, R2_NOT);

	hPen = CreatePen(PS_SOLID, 1, RGB(255,255,255));
	hPenPrev = SelectObject(hdc, hPen);
	hBrush = CreateSolidBrush(RGB(255,255,255));
	hBrushPrev = SelectObject(hdc, hBrush);

	for (i = 0; i < nVectors; i++) {
		if ((VectLst[i].f & 4) && ((sx != VectLst[i].x)||(sy != VectLst[i].y))) {
			sx = VectLst[i].x;
			sy = VectLst[i].y;
			if (mode & 4) {
				vx = (sx >> nZoom) - DispOfs.x + px;
				vy = (sy >> nZoom) - DispOfs.y + py;
				Ellipse(hdc, vx - 2, vy - 2, vx + 3, vy + 3);
			}
			vx = (sx >> nZoom) - DispOfs.x + x;
			vy = (sy >> nZoom) - DispOfs.y + y;
			Ellipse(hdc, vx - 2, vy - 2, vx + 3, vy + 3);
		}
	}

	px = x; py = y;

	SelectObject(hdc, hPenPrev);
	DeleteObject(hPen);
	SelectObject(hdc, hBrushPrev);
	DeleteObject(hBrush);

	SetROP2(hdc, R2_COPYPEN);
	ReleaseDC(hDrawWnd, hdc);

}




void DragSelecting (int mode, long sx, long sy, long ex, long ey)
{
	static long psx, psy, pex, pey;
	HDC hdc;
	HPEN hPen, hPenPrev;
	HBRUSH hBrushPrev;


	hdc = GetDC(hDrawWnd);
	SetROP2(hdc, R2_NOT);

	hPen = CreatePen(PS_SOLID, 1, RGB(255,255,255));
	hPenPrev = SelectObject(hdc, hPen);
	hBrushPrev = SelectObject(hdc, GetStockObject(NULL_BRUSH));

	if (mode & 4) {
		Rectangle(hdc, psx, psy, pex, pey);
	}
	Rectangle(hdc, sx, sy, ex, ey);

	psx = sx; psy = sy;
	pex = ex; pey = ey;

	SelectObject(hdc, hPenPrev);
	DeleteObject(hPen);
	SelectObject(hdc, hBrushPrev);

	SetROP2(hdc, R2_COPYPEN);
	ReleaseDC(hDrawWnd, hdc);

}




void DragSelectEnd (long sx, long sy, long ex, long ey)
{
	int i;
	long t;


	sx = (sx + DispOfs.x) << nZoom;
	sy = (sy + DispOfs.y) << nZoom;
	ex = (ex + DispOfs.x) << nZoom;
	ey = (ey + DispOfs.y) << nZoom;
	if (sx > ex) { t = sx; sx = ex; ex = t; }
	if (sy > ey) { t = sy; sy = ey; ey = t; }

	for (i = t = 0; i < nVectors; i++) {
		if ((VectLst[i].x >= sx) &&
			(VectLst[i].x <= ex) &&
			(VectLst[i].y >= sy) &&
			(VectLst[i].y <= ey) )
		{
				VectLst[i].f |= 2;
				t = 1;
		}
	}

	if (!t) {
		for (i = 0; i < nVectors; i++)
			VectLst[i].f &= 1;
	}
}




BOOL DragDragEnd(long x, long y)
{
	long i;


	x <<= nZoom; y <<= nZoom;

	for (i = 0; i < nVectors; i++) {
		if (VectLst[i].f & 4) {
			if ( ((VectLst[i].x + x) > 32767) ||
				((VectLst[i].x + x) < -32768) ) return FALSE;
			if ( ((VectLst[i].y + y) > 32767) ||
				((VectLst[i].y + y) < -32768) ) return FALSE;
		}
	}
	for (i = 0; i < nVectors; i++) {
		if (VectLst[i].f & 4) {
			VectLst[i].x += (short)x;
			VectLst[i].y += (short)y;
		}
	}
	return TRUE;
}




BOOL LoadCsvFile (HWND hWnd, char *fname, int mode)
{
	FILE *fp;
	char str[50], stc[16], c, d;
	int pps = 0, delay = 0, frm, res, n, p, q, rf, rx, ry;


	fp = fopen(fname, "rt");
	if (fp == NULL) {
		MessageBox(hWnd, "Failed to open frame file.", "Loading Frames", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

	if (fgets(str, sizeof(str), fp) != NULL) {
		if (sscanf(str, "VECTOR,%d,%d", &pps, &delay) >= 1) {
			if (mode == 0) {
				CurrentFrame = 0;
				for (n = 0; n < MAX_FRAMES; n++) {
					VectorsLst[n] = 0;
					memset(StrLst, ' ', sizeof(StrLst));
				}
				nFrames = 0;
				if (pps < 1000) pps = 1000;
				if (pps > 30000) pps = 30000;
				nPps = pps;
				if (delay < 0) delay = 0;
				if (delay >=32) delay = 32;
				nDelay = delay;
			}
			frm = CurrentFrame;
			do {
				if (InsertBlankFrame(frm) == FALSE) break;
				res = 3; n = 0;
				while (fgets(str, sizeof(str), fp) != NULL) {
					if (str[0] == '+') {
						res = 0;
						break;
					}
					res = sscanf(str, "%d,%d,%d",  &rf, &rx, &ry);
					if (res != 3) {
						p = 0; c = str[p];
						if (c == '"') p++;
						for (q = 0; q < 16; ) {
							d = str[p++];
							if (d < ' ') break;
							if ((d == '"') && (c == '"')) break;
							if ((d == ',') && (c != '"')) break;
							stc[q++] = d;
						}
						for (; q < 16; q++)
							stc[q] = ' ';
						memcpy(&StrLst[frm][0], stc, sizeof(stc));
						res = 3; continue;
					}
					if (n < MAX_VECTORS) {
						VectLst[n].f = (short)rf & 1;
						VectLst[n].x = (short)rx;
						VectLst[n].y = (short)-ry;
						n++;
					}
				}
				nVectors = n;
				PutFrame(frm, 1);
				frm++;
			} while (res != 3);

			fclose(fp);
			GetFrame(CurrentFrame);
			CreateUndo(2);
			return TRUE;
		}
	}

	fclose(fp);
	MessageBox(hWnd, "Invalid Csv frame file.", "Loading Frames", MB_OK | MB_ICONEXCLAMATION);
	return FALSE;

}




BOOL LoadIldFile (HWND hWnd, char *fname, int mode)
{
	FILE *fp;
	ILDA header;
	BYTE buff[8];
	int frm, n, s;
	UINT pl;


	fp = fopen(fname, "rb");
	if (fp == NULL) {
		MessageBox(hWnd, "Failed to open frame file.", "Loading Frames", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

	if (mode == 0) {
		CurrentFrame = 0;
		for (n = 0; n < MAX_FRAMES; n++) {
			VectorsLst[n] = 0;
			memset(StrLst, ' ', sizeof(StrLst));
		}
		nFrames = 0;
	}
	frm = CurrentFrame;

	while (fread(&header, 1, sizeof(ILDA), fp) == sizeof(ILDA)) {
		if (strcmp(&header.sign[0], "ILDA") != 0) break;
		if (header.points == 0) break;
		if (header.format > 2) break;
		if (header.format == 2) pl = 3;
		if (header.format == 1) pl = 6;
		else					   pl = 8;

		if (header.format <= 1) {
			if (InsertBlankFrame(frm) == FALSE) break;
			memcpy(&StrLst[frm][0], &header.framename[0], sizeof(StrLst[0]));
			for (n = s = 0; n < LEWORD(header.points); n++) {
				if (fread(buff, 1, pl, fp) != pl) break;
				if (s < MAX_VECTORS) {
					VectLst[s].x = (short)(((WORD)buff[0] << 8) + buff[1]);
					VectLst[s].y = -(short)(((WORD)buff[2] << 8) + buff[3]);
					VectLst[s].f = (buff[pl-2] & 0x40) ? 0 : 1;
					s++;
				}
			}
			nVectors = s;
			PutFrame(frm, 1);
			frm++;
		}
		else {
			for (n = s = 0; n < LEWORD(header.points); n++) {
				if (fread(buff, 1, pl, fp) != pl) break;
			}
		}
	}

	fclose(fp);
	GetFrame(CurrentFrame);
	CreateUndo(2);
	return TRUE;
}




BOOL LoadTraceImg (HWND hWnd, char *OpenName)
{
	OPENFILENAME lfn;
	char ImgFile[_MAX_PATH];
	HBITMAP hImg;
	HDC hdc;
	BITMAPINFOHEADER infoHeader;
	BITMAP infoImage;
	BYTE *ImgBuff;
	long n;
	BOOL f;


	if (OpenName == NULL) {
		ImgFile[0] = '\0';
		lfn.lStructSize			= sizeof(OPENFILENAME);
		lfn.Flags				= OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
		lfn.hwndOwner			= hWnd;
		lfn.lpstrTitle			= "Load Template Image";
		lfn.lpstrInitialDir		= NULL;
		lfn.lpstrFilter			= "Image Files (*.bmp)\0*.bmp\0All Files (*.*)\0*.*\0\0";
		lfn.lpstrCustomFilter	= NULL;
		lfn.nFilterIndex		= 0;
		lfn.lpstrDefExt			= "bmp";
		lfn.lpstrFile			= ImgFile;
		lfn.nMaxFile			= sizeof(ImgFile);
		lfn.lpstrFileTitle		= NULL;
		if (!GetOpenFileName(&lfn)) return FALSE;
	} else {
		strcpy(ImgFile, OpenName);
	}

	if ((hImg = (HBITMAP)LoadImage(NULL, ImgFile, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE)
		) == NULL) {
		MessageBox(hWnd, "Loading Failed.", "Load Image", MB_OK | MB_ICONWARNING);
		return FALSE;
	}
	if (hImage) DeleteObject(hImage);
	hImage = hImg;
	GetObject(hImage, sizeof(BITMAP), &infoImage);
	ImgSize.x = infoImage.bmWidth;
	ImgSize.y = infoImage.bmHeight;
	ImgBuff = (BYTE *)malloc((DWORD)ImgSize.x * (DWORD)ImgSize.y * 3);
	if (ImgBuff) {
		infoHeader.biSize = sizeof(BITMAPINFOHEADER);
		infoHeader.biWidth = ImgSize.x;
		infoHeader.biHeight = ImgSize.y;
		infoHeader.biPlanes = 1;
		infoHeader.biBitCount = 24;
		infoHeader.biCompression = 0;
		hdc = GetDC(hWnd);
		f = GetDIBits(hdc, hImage, 0, ImgSize.y, ImgBuff, (BITMAPINFO *)&infoHeader, DIB_RGB_COLORS);
		for (n = 0; n < ImgSize.x * ImgSize.y * 3; n++)
			ImgBuff[n] /= 3;
		f = SetDIBits(hdc, hImage, 0, ImgSize.y, ImgBuff, (BITMAPINFO *)&infoHeader, DIB_RGB_COLORS);
		ReleaseDC(hWnd, hdc);
		free(ImgBuff);
	}

	EnableMenuItem(GetMenu(hMainWnd), IDM_CLRPIC, MF_BYCOMMAND | MF_ENABLED);
	return TRUE;
}




BOOL SaveCsvFile(HWND hWnd, char *fname, int n_from, int n_to)
{
	FILE *fp;
	int n, frm;
	char str[20];


	fp = fopen(fname, "wt");
	if (fp == NULL) {
		MessageBox(hWnd, "Failed to create frame file.", "Saving Frames", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

	fprintf(fp, "VECTOR,%d,%d\n", nPps, nDelay);
	for (frm = n_from; frm <= n_to; frm++){
		memcpy(str, &StrLst[frm][0], sizeof(StrLst[0]));
		str[16] = 0;
		if (strstr(str, ","))
			fprintf(fp, "\"%s\"\n", str);
		else
			fprintf(fp, "%s\n", str);
		for (n = 0; n < VectorsLst[frm]; n++)
			fprintf(fp, "%d,%d,%d\n", FrameLst[frm][n].f & 1, FrameLst[frm][n].x, -FrameLst[frm][n].y);
		if (frm < n_to)
			fprintf(fp, "+\n");
	}

	fclose(fp);
	return TRUE;
}




BOOL SaveIldaFile(HWND hWnd, char *fname, int n_from, int n_to)
{
	FILE *fp;
	int n, frm;
	ILDA header;
	WORD vec[4];


	for (n = n_from; n <= n_to; n++) {
		if (VectorsLst[n] == 0) {
			MessageBox(hWnd, "ILDA format does not allow any blank frame.", "Saving Frames", MB_OK | MB_ICONEXCLAMATION);
			return FALSE;
		}
	}

	fp = fopen(fname, "wb");
	if (fp == NULL) {
		MessageBox(hWnd, "Failed to create frame file.", "Saving Frames", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

	memset(&header, 0, sizeof(ILDA));
	memcpy(&header.sign[0], "ILDA\0\0\0", 7);
	header.totalframes = LEWORD(n_to - n_from + 1);
	vec[2] = 0;

	for (frm = n_from; frm <= n_to; frm++) {
		memcpy(&header.framename[0], &StrLst[frm], 16);
		header.points = LEWORD(VectorsLst[frm]);
		fwrite(&header, 1, sizeof(header), fp);
		for (n = 0; n < VectorsLst[frm]; n++) {
			vec[0] = LEWORD(FrameLst[frm][n].x);
			vec[1] = LEWORD(-FrameLst[frm][n].y);
			vec[3] = LEWORD((FrameLst[frm][n].f & 1) ? 0x0001 : 0x4000);
			fwrite(vec, 1, sizeof(vec), fp);
		}
		header.framenum = LEWORD(LEWORD(header.framenum) + 1);
	}
	header.points = 0;
	fwrite(&header, 1, sizeof(header), fp);

	fclose(fp);
	return TRUE;
}




BOOL SaveFrmFile(HWND hWnd, char *fname, int n_from, int n_to)
{
	FILE *fp;
	int n, frm;
	DWORD sreg = 0;
	WORD head[16];


	fp = fopen(fname, "wb");
	if (fp == NULL) {
		MessageBox(hWnd, "Failed to create frame file.", "Saving Frames", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

	memcpy(&head[0], "VD  ", 4);
	head[2] = nPps;
	for (frm = n_from, n = 0; frm <= n_to; frm++)
		n += VectorsLst[frm];
	*(int *)(&head[3]) = n;
	memset(&head[5], 0, 6);
	memcpy(&head[8], &StrLst[n_from], 16);
	fwrite(head, 1, sizeof(head), fp);

	for (frm = n_from; frm <= n_to; frm++) {
		for (n = 0; n < VectorsLst[frm]; n++) {
			head[0] = (WORD)(FrameLst[frm][n].x - 32768) >> 4;
			head[1] = (WORD)(-FrameLst[frm][n].y - 32768) >> 4;
			sreg <<= 1;
			if (FrameLst[frm][n].f & 1) sreg++;
			if (sreg & (1 << nDelay)) head[0] |= 0x8000;
			fwrite(head, 1, 4, fp);
		}
	}

	fclose(fp);
	return TRUE;
}




BOOL SaveFrames (HWND hWnd,		/* Window handle */
				int mode,		/* b0: Enter file name, b1: Don't change current name */
				int fn_from,	/* Start frame# to be written */
				int fn_to		/* End frame# to be written */
				)
{
	OPENFILENAME lfn;
	char SaveFile[_MAX_PATH], FileTitle[_MAX_FNAME];
	int n;
	BOOL result = FALSE;


	strcpy(SaveFile, (mode & 2) ? "" : FileName);

	if ((mode & 1) || !strlen(SaveFile)) {
		lfn.lStructSize			= sizeof(OPENFILENAME);
		lfn.Flags				= OFN_OVERWRITEPROMPT | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY | OFN_ENABLEHOOK | OFN_EXPLORER;
		lfn.hwndOwner			= hWnd;
		lfn.lpfnHook			= OpenFileProc;
		lfn.lpstrTitle			= (mode & 2) ? "Write Frames" : "Save Frames";
		lfn.lpstrInitialDir		= NULL;
		lfn.lpstrFilter			= "Csv Files (*.csv)\0*.csv\0ILDA Files (*.ild)\0*.ild\0Packed files (*.frm)\0*.frm\0All Files (*.*)\0*.*\0";
		lfn.lpstrCustomFilter	= NULL;
		lfn.nFilterIndex		= FilterIdx;
		lfn.lpstrDefExt			= sFileExt[FilterIdx];
		lfn.lpstrFile			= SaveFile;
		lfn.nMaxFile			= sizeof(SaveFile);
		lfn.lpstrFileTitle		= FileTitle;
		lfn.nMaxFileTitle		= sizeof(FileTitle);
		if (!GetSaveFileName(&lfn)) return FALSE;
		mode |= 1;
	}

	if ((n = strlen(SaveFile)) < 4) return FALSE;

	if (_stricmp(&SaveFile[n-4], ".CSV") == 0)
		result = SaveCsvFile(hWnd, SaveFile, fn_from, fn_to);
	if (_stricmp(&SaveFile[n-4], ".ILD") == 0)
		result = SaveIldaFile(hWnd, SaveFile, fn_from, fn_to);
	if (_stricmp(&SaveFile[n-4], ".FRM") == 0) {
		SaveFrmFile(hWnd, SaveFile, fn_from, fn_to);
		return FALSE;
	}

	if (result && (mode == 1)) {
		FilterIdx = lfn.nFilterIndex;
		strcpy(FileName, SaveFile);
		strcpy(DispName, FileTitle);
	}
	return result;
}




BOOL OpenFrame (HWND hWnd,			/* Window Handle */
				char *OpenName,		/* File Name */
				int mode			/* 0:Load as new, 1:Insert frames */
				)
{
	OPENFILENAME lfn;
	char OpenFile[_MAX_PATH], FileTitle[_MAX_FNAME];
	int n;
	BOOL result, name_entered = FALSE;


	if (OpenName == NULL) {	/* Open file dialog when no name given */
		OpenFile[0] = '\0';
		lfn.lStructSize			= sizeof(OPENFILENAME);
		lfn.Flags				= OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
		lfn.hwndOwner			= hWnd;
		lfn.lpstrTitle			= mode ? "Insert Frames" : "Open Frames";
		lfn.lpstrInitialDir		= NULL;
		lfn.lpstrFilter			= "Csv Files (*.csv)\0*.csv\0ILDA Files (*.ild)\0*.ild\0All Files (*.*)\0*.*\0";
		lfn.lpstrCustomFilter	= NULL;
		lfn.nFilterIndex		= FilterIdx;
		lfn.lpstrDefExt			= sFileExt[FilterIdx];
		lfn.lpstrFile			= OpenFile;
		lfn.nMaxFile			= sizeof(OpenFile);
		lfn.lpstrFileTitle		= FileTitle;
		lfn.nMaxFileTitle		= sizeof(FileTitle);
		if (!GetOpenFileName(&lfn)) return FALSE;
		name_entered = TRUE;
	} else {
		strcpy(OpenFile, OpenName);
	}

	if ((n = strlen(OpenFile)) < 4) return FALSE;

	if (_stricmp(&OpenFile[n-4], ".CSV") == 0) {
		result = LoadCsvFile(hWnd, OpenFile, mode);
	} else {
		if (_stricmp(&OpenFile[n-4], ".ILD") == 0) {
			result = LoadIldFile(hWnd, OpenFile, mode);
		} else {
			return FALSE;
		}
	}

	if (result && (mode == 0)) {
		if (name_entered) FilterIdx = lfn.nFilterIndex;
		strcpy(FileName, OpenFile);
		strcpy(DispName, FileTitle);
	}

	return result;
}




LRESULT CALLBACK
VectWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	RECT crect;
	LVCOLUMN cols;
	HMENU hmnu1, hmnu2;
	POINT csr;
	int i;


	switch (msg) {
		case WM_CREATE :
			GetClientRect(hWnd, &crect);
			hVectViewWnd = CreateWindowEx(WS_EX_CLIENTEDGE,
										WC_LISTVIEW, "",
										WS_CHILD | WS_VISIBLE | LVS_REPORT,
										0, 0, crect.right + 1, crect.bottom + 1,
										hWnd, NULL, hMainInst, NULL);
			if (!hVectViewWnd) return -1;
			cols.mask = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH;
			cols.fmt = LVCFMT_RIGHT;
			cols.cx = 36;
			cols.pszText = "No.";
			ListView_InsertColumn(hVectViewWnd, 0, &cols);
			cols.cx = 22;
			cols.pszText = "V";
			ListView_InsertColumn(hVectViewWnd, 1, &cols);
			cols.cx = 55;
			cols.pszText = "X";
			ListView_InsertColumn(hVectViewWnd, 2, &cols);
			cols.pszText = "Y";
			ListView_InsertColumn(hVectViewWnd, 3, &cols);
			ListView_SetExtendedListViewStyle(hVectViewWnd, ListView_GetExtendedListViewStyle(hVectViewWnd)
															| LVS_EX_FULLROWSELECT
															| LVS_EX_GRIDLINES);
			break;

		case WM_SIZE :
			MoveWindow(hVectViewWnd, 0, 0,
						LOWORD(lParam) + 1,
						HIWORD(lParam) + 1,
						TRUE);
			break;

		case WM_COMMAND :
			switch (LOWORD(wParam)) {
				case IDM_LSTINS :
					LV_InsertVector();
					break;
				case IDM_LSTDEL :
					LV_DeleteVectors();
					break;
				case IDM_LSTEDT :
					LV_EditVector(hWnd);
					break;
			}
			break;

		case WM_NOTIFY :
			switch(((LPNMHDR)lParam)->code) {
				case LVN_ITEMCHANGED :
					if ((((LPNMLISTVIEW)lParam)->uChanged & LVIF_STATE) &&
						((i = ((LPNMLISTVIEW)lParam)->iItem) >= 0)) {
						if (ListView_GetItemState(hVectViewWnd, i, LVIS_SELECTED))
							VectLst[i].f |= 2;
						else
							VectLst[i].f &= ~2;
						InvalidateRect(hDrawWnd, NULL, TRUE);
					}
					break;
				case LVN_KEYDOWN :
					switch (((LPNMLVKEYDOWN)lParam)->wVKey) {
						case VK_INSERT :
							LV_InsertVector();
							break;
						case VK_DELETE :
							LV_DeleteVectors();
							break;
						case VK_RETURN :
							LV_EditVector(hWnd);
							break;
					}
					break;
				case NM_DBLCLK :
					LV_EditVector(hWnd);
					break;
				case NM_RCLICK :
					if (hmnu1 = LoadMenu(hMainInst, MAKEINTRESOURCE(IDR_MENU2))) {
						if (hmnu2 = GetSubMenu(hmnu1, 0)) {
							GetCursorPos(&csr);
							TrackPopupMenu(hmnu2, TPM_LEFTALIGN | TPM_TOPALIGN,
											csr.x, csr.y, 0, hWnd, NULL);
						}
						DestroyMenu(hmnu1);
					}
					break;
			}
			break;

		case WM_SETFOCUS :
			SetFocus(hVectViewWnd);
			break;

		case WM_DESTROY :
			hVectWnd = NULL;

		default :
			return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}




LRESULT CALLBACK
FrameWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	RECT crect;
	LVCOLUMN cols;
	HMENU hmnu1, hmnu2;
	POINT csr;
	int i;


	switch (msg) {
		case WM_CREATE :
			GetClientRect(hWnd, &crect);
			hFrameViewWnd = CreateWindowEx(WS_EX_CLIENTEDGE,
										WC_LISTVIEW, "",
										WS_CHILD | WS_VISIBLE | LVS_REPORT,
										0, 0, crect.right + 1, crect.bottom + 1,
										hWnd, NULL, hMainInst, NULL);
			if (!hFrameViewWnd) return -1;
			cols.mask = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH;
			cols.fmt = LVCFMT_RIGHT;
			cols.cx = 36;
			cols.pszText = "Frm";
			ListView_InsertColumn(hFrameViewWnd, 0, &cols);
			cols.fmt = LVCFMT_LEFT;
			cols.cx = 150;
			cols.pszText = "Name";
			ListView_InsertColumn(hFrameViewWnd, 1, &cols);
			ListView_SetExtendedListViewStyle(hFrameViewWnd, ListView_GetExtendedListViewStyle(hFrameViewWnd)
															| LVS_EX_FULLROWSELECT
															| LVS_EX_GRIDLINES);
			break;

		case WM_SIZE :
			MoveWindow(hFrameViewWnd, 0, 0,
						LOWORD(lParam) + 1,
						HIWORD(lParam) + 1,
						TRUE);
			break;

		case WM_COMMAND :
			switch (LOWORD(wParam)) {
				case IDM_LSTEDT :
					FV_EditFrame(hWnd);
					break;
				case IDM_LSTINS :
					FV_InsertFrame();
					break;
				case IDM_LSTDEL :
					FV_DeleteFrames();
					break;
				case IDM_LSTRESIZE :
					FV_ResizeFrames(hWnd);
					break;
			}
			break;

		case WM_NOTIFY :
			switch(((LPNMHDR)lParam)->code) {
				case LVN_ITEMCHANGED :
					i = ListView_GetNextItem(hFrameViewWnd, -1, LVNI_FOCUSED);
					if (i >= 0) ChangeFrame(i);
					break;
				case LVN_KEYDOWN :
					switch (((LPNMLVKEYDOWN)lParam)->wVKey) {
						case VK_INSERT :
							FV_InsertFrame();
							break;
						case VK_DELETE :
							FV_DeleteFrames();
							break;
						case VK_RETURN :
							FV_EditFrame(hWnd);
							break;
					}
					break;
				case NM_DBLCLK :
					FV_EditFrame(hWnd);
					break;
				case NM_RCLICK :
					i = ListView_GetNextItem(hFrameViewWnd, -1, LVNI_FOCUSED);
					if (i >= 0) ChangeFrame(i);
					if (hmnu1 = LoadMenu(hMainInst, MAKEINTRESOURCE(IDR_MENU3))) {
						if (hmnu2 = GetSubMenu(hmnu1, 0)) {
							GetCursorPos(&csr);
							TrackPopupMenu(hmnu2, TPM_LEFTALIGN | TPM_TOPALIGN,
											csr.x, csr.y, 0, hWnd, NULL);
						}
						DestroyMenu(hmnu1);
					}
					break;
			}
			break;

		case WM_SETFOCUS :
			SetFocus(hFrameViewWnd);
			break;

		case WM_DESTROY :
			hFrameWnd = NULL;

		default :
			return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}




void RedrawSheet(HWND hWnd)
{
	HDC hdc, hdc_img;
	HPEN hPenV, hPenI, hPenL, hPenF, hPenS, hPenPrev;
	HBRUSH hBrushT, hBrushF, hBrushPrev;
	RECT crect;
	PAINTSTRUCT pnt;
	SCROLLINFO scrl;
	LONG x, y, px, py, sx, sy, wx, wy;
	int i;
	int framesize = (65536 >> nZoom);
	int frameover = (67000 >> nZoom);


	GetClientRect(hWnd, &crect);
	wx = crect.right + 1;
	wy = crect.bottom + 1;

	if (fImgLoaded) {
		DispOfs.x = wx / -2;
		DispOfs.y = wy / -2;
		fImgLoaded = FALSE;
	}

	scrl.cbSize = sizeof(SCROLLINFO);
	scrl.nMin = 0;
	scrl.fMask = SIF_DISABLENOSCROLL | SIF_PAGE | SIF_POS | SIF_RANGE;

	if (wx >= frameover) {			/* Client size is equal or larger than frame size */
		DispOfs.x = wx / -2;
		scrl.nMax = 99;
		scrl.nPage = 100;
	} else {								/* Client size is smaller than picture size */
		if (frameover / 2 + DispOfs.x < 0)	/* There is a space left of picture */
			DispOfs.x = frameover / -2;
		if (frameover / 2 - DispOfs.x < wx)	/* There is a space right of picture */
			DispOfs.x = frameover / 2 - wx;
		scrl.nMax	= frameover - 1;
		scrl.nPage	= wx;
		scrl.nPos	= frameover / 2 + DispOfs.x;
	}
	SetScrollInfo(hWnd, SB_HORZ, &scrl, TRUE);

	if (wy >= frameover) {					/* Client size is equal or larger than picture size */
		DispOfs.y = wy / -2;
		scrl.nMax = 99;
		scrl.nPage = 100;
	} else {								/* Client size if smaller than picture size */
		if (frameover / 2 + DispOfs.y < 0)	/* There is a space top of picture */
			DispOfs.y = frameover / -2;
		if (frameover / 2 - DispOfs.y < wy)	/* There is a space bottom of picture */
			DispOfs.y = frameover / 2 - wy;
		scrl.nMax	= frameover - 1;
		scrl.nPage	= wy;
		scrl.nPos	= frameover / 2 + DispOfs.y;
	}
	SetScrollInfo(hWnd, SB_VERT, &scrl, TRUE);

	hdc = BeginPaint(hWnd, &pnt);

	if (hImage) {
		if (ImgSize.x / 2 + DispOfs.x < 0) {	/* There is a space left of picture */
			x = ImgSize.x / -2 - DispOfs.x; sx = 0;
		} else {								/* Picture left is out of window */
			x = 0; sx = ImgSize.x / 2 + DispOfs.x;
		}
		if (ImgSize.y / 2 + DispOfs.y < 0) {	/* There is a space top of picture */
			y = ImgSize.y / -2 - DispOfs.y; sy = 0;
		} else {								/* Picture top is out of window */
			y = 0; sy = ImgSize.y / 2 + DispOfs.y;
		}
		if (ImgSize.x / 2 - DispOfs.x <= wx) px = ImgSize.x; /* Clip right of image if needed */
		else px = wx + DispOfs.x + ImgSize.x / 2;
		if (ImgSize.y / 2 - DispOfs.y <= wy) py = ImgSize.y;  /* Clip bottom of image if needed */
		else py = wy + DispOfs.y + ImgSize.y / 2;

		if ((px > 0) && (py > 0) && (sx < ImgSize.x) && (sy < ImgSize.y)) {
			hdc_img = CreateCompatibleDC(hdc);
			SelectObject(hdc_img, hImage);
			BitBlt(hdc, x, y, px, py, hdc_img, sx, sy, SRCCOPY);
			DeleteDC(hdc_img);
		}
	}

	hPenV = CreatePen(PS_SOLID, 1, RGB(0,255,0));	// Green (visible vector)
	hPenI = CreatePen(PS_SOLID, 1, RGB(0,0,255));	// Blue (invisible vector)
	hPenF = CreatePen(PS_SOLID, 1, RGB(128,0,0));	// Dark-Red (outline)
	hPenS = CreatePen(PS_SOLID, 1, RGB(255,0,0));	// Red (selected vector)
	hPenL = CreatePen(PS_DOT, 1, RGB(255,255,255));	// Dotted-White (center cross)
	hBrushT = GetStockObject(NULL_BRUSH);			// Transparent (inside vector)
	hBrushF = CreateSolidBrush(RGB(255,0,0));		// Red (inside selected vector)
	hBrushPrev = SelectObject(hdc, hBrushT);
	SetBkMode(hdc, TRANSPARENT);

	/* Outline */
	hPenPrev = SelectObject(hdc, hPenF);
	Rectangle(hdc, -framesize / 2 - DispOfs.x, -framesize / 2 - DispOfs.y,
					framesize / 2 - DispOfs.x, framesize / 2 - DispOfs.y);

	/* Center Cross */
	SelectObject(hdc, hPenL);
	MoveToEx(hdc, -framesize / 2 - DispOfs.x, -DispOfs.y, NULL);
	LineTo(hdc, framesize / 2 - DispOfs.x, -DispOfs.y);
	MoveToEx(hdc, -DispOfs.x, -framesize / 2 - DispOfs.y, NULL);
	LineTo(hdc, -DispOfs.x, framesize / 2 - DispOfs.y);

	/* Vectors */
	for (i = 0; i < nVectors; i++) {
		x = (VectLst[i].x >> nZoom) - DispOfs.x;
		y = (VectLst[i].y >> nZoom) - DispOfs.y;
		if (VectLst[i].f & 1)
			SelectObject(hdc, hPenV);
		else
			SelectObject(hdc, hPenI);
		if (i) {	// 2nd vector and followings...
			if (((VectLst[i].f & 1) != (VectLst[i-1].f & 1)) ||	// skip duplicated vects
				(VectLst[i].x != VectLst[i-1].x) ||
				(VectLst[i].y != VectLst[i-1].y)) {
				LineTo(hdc, x, y);
				Ellipse(hdc, x - 2, y - 2, x + 3, y + 3);
			}
		} else {	// 1st vector
			MoveToEx(hdc, x, y, NULL);
			Ellipse(hdc, x - 2, y - 2, x + 3, y + 3);
		}
	}

	/* Selections */
	SelectObject(hdc, hPenS);
	SelectObject(hdc, hBrushF);
	px = py = 100000;
	for (i = 0; i < nVectors; i++) {
		if (VectLst[i].f & 2) {
			x = (VectLst[i].x >> nZoom) - DispOfs.x;
			y = (VectLst[i].y >> nZoom) - DispOfs.y;
			if ((px != x)||(py != y)) {
				Ellipse(hdc, x - 2, y - 2, x + 3, y + 3);
				px = x; py = y;
			}
		}
	}

	SelectObject(hdc, hBrushPrev);
	DeleteObject(hBrushF);
	SelectObject(hdc, hPenPrev);
	DeleteObject(hPenI);
	DeleteObject(hPenV);
	DeleteObject(hPenF);
	DeleteObject(hPenL);
	DeleteObject(hPenS);

	EndPaint(hWnd, &pnt);

}




LRESULT CALLBACK
DrawWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	char str[30];
	SCROLLINFO scrl;
	static int drag = 0;
	static POINT ds;
	RECT crect;


	switch (msg) {
		case WM_LBUTTONDOWN :
			if ((wParam & MK_SHIFT)&&(wParam & MK_CONTROL)) {
				ds.x = (long)LOWORD(lParam);
				ds.y = (long)HIWORD(lParam);
				drag = 2;
				break;
			}
			if (wParam & MK_CONTROL) {
				ds.x = (long)LOWORD(lParam);
				ds.y = (long)HIWORD(lParam);
				if (DragDragStart(ds.x, ds.y))
					drag = 1;
				break;
			}
			if (!AddVector((wParam & MK_SHIFT) ? 0 : 1, LOWORD(lParam), HIWORD(lParam))) {
				MessageBeep(MB_OK);
				break;
			}
			SetChanged(TRUE);
			UpdateCaption(2);
			break;

		case WM_LBUTTONUP :
			if (drag) {
				if (drag == 5) {
					CreateUndo(1);
					if (DragDragEnd((long)LOWORD(lParam) - ds.x, (long)HIWORD(lParam) - ds.y)) {
						LV_RfshVector(2);
						SetChanged(TRUE);
					} else {
						MessageBeep(MB_OK);
					}
				}
				if (drag & 2) {
					DragSelectEnd(ds.x, ds.y, (long)LOWORD(lParam), (long)HIWORD(lParam));
					LV_RfshVector(2);
				}
				drag = 0;
				InvalidateRect(hWnd, NULL, TRUE);
			}
			break;

		case WM_RBUTTONDOWN :
			if (drag) {
				drag = 0;
				InvalidateRect(hWnd, NULL, TRUE);
				break;
			}
			if (wParam & MK_CONTROL) break;
			if (wParam & MK_SHIFT) break;
			if (SubtractVector(LOWORD(lParam), HIWORD(lParam))) {
				SetChanged(TRUE);
				UpdateCaption(2);
				InvalidateRect(hWnd, NULL, TRUE);
			}
			break;

		case WM_MOUSEMOVE :
			if (drag) {
				if (drag & 2)
					DragSelecting(drag, ds.x, ds.y, (long)LOWORD(lParam), (long)HIWORD(lParam));
				else
					DragDragging(drag, (long)LOWORD(lParam) - ds.x, (long)HIWORD(lParam) - ds.y);
				drag |= 4;
			}
			sprintf(str, "%d, %d", (DispOfs.x + LOWORD(lParam)) << nZoom, (-(DispOfs.y + HIWORD(lParam)) << nZoom));
			SendMessage(hSBWnd, SB_SETTEXT, 3, (LPARAM)str);
			break;

		case WM_HSCROLL :
			scrl.cbSize		= sizeof(SCROLLINFO);
			scrl.fMask		= SIF_ALL;
			GetScrollInfo(hDrawWnd, SB_HORZ, &scrl);
			GetClientRect(hDrawWnd, &crect);
			switch (LOWORD(wParam)) {
				case SB_LINELEFT :
					DispOfs.x -= crect.right / 16;
					break;
				case SB_LINERIGHT :
					DispOfs.x += crect.right / 16;
					break;
				case SB_PAGELEFT :
					DispOfs.x -= crect.right / 2;
					break;
				case SB_PAGERIGHT :
					DispOfs.x += crect.right / 2;
					break;
				case SB_THUMBPOSITION :
					DispOfs.x = scrl.nTrackPos - (scrl.nMax + 1) / 2; 
					break;
				default :
					return 0;
			}
			InvalidateRect(hWnd, NULL, TRUE);
			break;

		case WM_VSCROLL :
			scrl.cbSize		= sizeof(SCROLLINFO);
			scrl.fMask		= SIF_ALL;
			GetScrollInfo(hDrawWnd, SB_VERT, &scrl);
			GetClientRect(hDrawWnd, &crect);
			switch (LOWORD(wParam)) {
				case SB_LINELEFT :
					DispOfs.y -= crect.bottom / 16;
					break;
				case SB_LINERIGHT :
					DispOfs.y += crect.bottom / 16;
					break;
				case SB_PAGELEFT :
					DispOfs.y -= crect.bottom / 2;
					break;
				case SB_PAGERIGHT :
					DispOfs.y += crect.bottom / 2;
					break;
				case SB_THUMBPOSITION :
					DispOfs.y = scrl.nTrackPos - (scrl.nMax + 1) / 2; 
					break;
				default :
					return 0;
			}
			InvalidateRect(hWnd, NULL, TRUE);
			break;

		case WM_PAINT :
			RedrawSheet(hWnd);
			break;

		default :
			return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}




LRESULT CALLBACK
MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	int sbSizes[4] = {180, 270, 340, 430};
	int res, f_from, f_to;
	INITCOMMONCONTROLSEX comex;
	RECT crect;
	POINT cpnt;
	char DropFile[_MAX_FNAME];


	switch (msg) {
		case WM_COMMAND :
			switch (LOWORD(wParam)) {
				case IDM_EXIT :
					SendMessage(hWnd, WM_CLOSE, 0, 0);
					break;

				case IDM_ABOUT :
					MessageBox(hWnd, VerStr, "Version", MB_OK);
					break;

				case IDM_HELP :
					ShellExecute(hWnd, NULL, "mkv2.html", NULL, NULL, SW_SHOW);
					break;

				case IDM_APPEND :
					if (OpenFrame(hWnd, NULL, 1)) {
						SetChanged(TRUE);
						InvalidateRect(hDrawWnd, NULL, TRUE);
						UpdateCaption(2);
						FV_RfshFrame(1);
						LV_RfshVector(0);
					}
					break;

				case IDM_OPEN :
					PutFrame(CurrentFrame, 0);
					if (fFileChanged) {
						res = MessageBox(hWnd, 
								"The changes have not been saved.\nDo you wish to save changes?", "Open Frame",
								MB_YESNOCANCEL | MB_ICONWARNING);
						if (res == IDCANCEL) return 0;
						if ((res == IDYES) && !SaveFrames(hWnd, 0, 0, nFrames-1)) return 0;
					}
					if (OpenFrame(hWnd, NULL, 0)) {
						SetChanged(FALSE);
						InvalidateRect(hDrawWnd, NULL, TRUE);
						UpdateCaption(3);
						FV_RfshFrame(1);
						LV_RfshVector(0);
					}
					break;

				case IDM_SAVE :
					PutFrame(CurrentFrame, 0);
					if (SaveFrames(hWnd, 0, 0, nFrames-1))
						SetChanged(FALSE);
						UpdateCaption(1);
					break;

				case IDM_SAVEAS :
					PutFrame(CurrentFrame, 0);
					if (SaveFrames(hWnd, 1, 0, nFrames-1)) {
						SetChanged(FALSE);
						UpdateCaption(1);
					}
					break;

				case IDM_WRITEFRAME :
					if (!FV_GetSelection(&f_from, &f_to)) {
						f_from = f_to = CurrentFrame;
					}
					PutFrame(CurrentFrame, 0);
					SaveFrames(hWnd, 2, f_from, f_to);
					break;

				case IDM_LDPIC :
					if (LoadTraceImg(hWnd, NULL)) {
						fImgLoaded = TRUE;
						InvalidateRect(hDrawWnd, NULL, TRUE);
					}
					break;

				case IDM_CLRPIC :
					if (hImage) DeleteObject(hImage);
					hImage = NULL;
					EnableMenuItem(GetMenu(hMainWnd), IDM_CLRPIC, MF_BYCOMMAND | MF_GRAYED);
					InvalidateRect(hDrawWnd, NULL, TRUE);
					break;

				case IDM_UNDO :
					if (Undo()) {
						SetChanged(TRUE);
						UpdateCaption(2);
						LV_RfshVector(0);
						InvalidateRect(hDrawWnd, NULL, TRUE);
					}
					break;

				case IDM_INSFRM :
					PutFrame(CurrentFrame, 0);
					if (InsertBlankFrame(CurrentFrame)) {
						GetFrame(CurrentFrame);
						SetChanged(TRUE);
						CreateUndo(2);
						UpdateCaption(2);
						FV_RfshFrame(1);
						LV_RfshVector(1);
						InvalidateRect(hDrawWnd, NULL, TRUE);
					}
					break;

				case IDM_DELFRM :
					if (DeleteFrame(CurrentFrame)) {
						GetFrame(CurrentFrame);
						SetChanged(TRUE);
						CreateUndo(2);
						UpdateCaption(2);
						FV_RfshFrame(1);
						LV_RfshVector(1);
						InvalidateRect(hDrawWnd, NULL, TRUE);
					}
					break;

				case IDM_RESIZE :
					if (CurrentFrame >= nFrames) break;
					CreateUndo(1);
					DlgCaption[0] = "Resize Frames";
					DlgCaption[1] = "Change frame size to:";
					DlgMode = 0;
					PutFrame(CurrentFrame, 0);
					_itoa(GetFrameSize(CurrentFrame, CurrentFrame), DlgStr1, 10);
					if (DialogBox(hMainInst, MAKEINTRESOURCE(IDD_OPTION), hWnd, (DLGPROC)OptWndProc)
						== IDOK) {
						ResizeFrames(atoi(DlgStr1), CurrentFrame, CurrentFrame);
						GetFrame(CurrentFrame);
						SetChanged(TRUE);
						LV_RfshVector(2);
						InvalidateRect(hDrawWnd, NULL, TRUE);
					}
					break;

				case IDM_SPEED :
					DlgCaption[0] = "Change Speed";
					DlgCaption[1] = "Enter scan speed [pps]:";
					DlgMode = 0;
					_itoa(nPps, DlgStr1, 10);
					if (DialogBox(hMainInst, MAKEINTRESOURCE(IDD_OPTION), hWnd, (DLGPROC)OptWndProc)
						== IDOK) {
						nPps = atoi(DlgStr1);
						if (nPps < 1000) nPps = 1000;
						if (nPps > 30000) nPps = 30000;
						SetChanged(TRUE);
						UpdateCaption(2);
					}
					break;

				case IDM_DELAY :
					DlgCaption[0] = "Change Delay";
					DlgCaption[1] = "Enter blanking delay [pts]:";
					DlgMode = 0;
					_itoa(nDelay, DlgStr1, 10);
					if (DialogBox(hMainInst, MAKEINTRESOURCE(IDD_OPTION), hWnd, (DLGPROC)OptWndProc)
						== IDOK) {
						nDelay = atoi(DlgStr1);
						if (nDelay < 0) nDelay = 0;
						if (nDelay > 30) nDelay = 30;
						SetChanged(TRUE);
						UpdateCaption(2);
					}
					break;

				case IDM_CLRVEC :
					if (nVectors) {
						CreateUndo(0);
						nVectors = 0;
						SetChanged(TRUE);
						UpdateCaption(2);
						LV_RfshVector(0);
						InvalidateRect(hDrawWnd, NULL, TRUE);
					}
					break;

				case IDM_NEXTFRM :
					if (CurrentFrame < MAX_FRAMES)
						ChangeFrame(CurrentFrame + 1);
					break;

				case IDM_PREVFRM :
					if (CurrentFrame > 0)
						ChangeFrame(CurrentFrame - 1);
					break;

				case IDM_SHOWLIST :
					if (!hVectWnd) {
						cpnt.x = cpnt.y = 0;
						ClientToScreen(hWnd, &cpnt);
						hVectWnd = CreateWindowEx(WS_EX_TOOLWINDOW | WS_EX_WINDOWEDGE,
												"mkv_vect", "Vector Table",
												WS_SYSMENU | WS_THICKFRAME | WS_VISIBLE,
												cpnt.x, cpnt.y, 202, 480,
												hWnd, NULL, hMainInst, NULL);
						LV_RfshVector(0);
					} else {
						DestroyWindow(hVectWnd);
					}
					break;

				case IDM_SHOWFRAME :
					if (!hFrameWnd) {
						cpnt.x = cpnt.y = 0;
						ClientToScreen(hWnd, &cpnt);
						hFrameWnd = CreateWindowEx(WS_EX_TOOLWINDOW | WS_EX_WINDOWEDGE,
												"mkv_frame", "Frame Table",
												WS_SYSMENU | WS_THICKFRAME | WS_VISIBLE,
												cpnt.x, cpnt.y, 220, 250,
												hWnd, NULL, hMainInst, NULL);
						FV_RfshFrame(1);
					} else {
						DestroyWindow(hFrameWnd);
					}
					break;

				case IDM_ZOOMOUT :
					if (nZoom < 6) {
						GetClientRect(hDrawWnd, &crect);
						cpnt.x = (DispOfs.x + (crect.right + 1) / 2) << nZoom;
						cpnt.y = (DispOfs.y + (crect.bottom + 1) / 2) << nZoom;
						nZoom++;
						DispOfs.x = (cpnt.x >> nZoom) - (crect.right + 1) / 2;
						DispOfs.y = (cpnt.y >> nZoom) - (crect.bottom + 1) / 2;
						UpdateCaption(2);
						InvalidateRect(hDrawWnd, NULL, TRUE);
					}
					break;

				case IDM_ZOOMIN :
					if (nZoom > 0) {
						GetClientRect(hDrawWnd, &crect);
						cpnt.x = (DispOfs.x + (crect.right + 1) / 2) << nZoom;
						cpnt.y = (DispOfs.y + (crect.bottom + 1) / 2) << nZoom;
						nZoom--;
						DispOfs.x = (cpnt.x >> nZoom) - (crect.right + 1) / 2;
						DispOfs.y = (cpnt.y >> nZoom) - (crect.bottom + 1) / 2;
						UpdateCaption(2);
						InvalidateRect(hDrawWnd, NULL, TRUE);
					}
					break;
			}
			break;

		case WM_DROPFILES :
			SetForegroundWindow(hWnd);
			DragQueryFile((HDROP)wParam, 0, DropFile, sizeof(DropFile));
			if (OpenFrame(hWnd, DropFile, 1)) {
				SetChanged(TRUE);
				InvalidateRect(hDrawWnd, NULL, TRUE);
				UpdateCaption(2);
				LV_RfshVector(0);
				FV_RfshFrame(1);
			} else {
				if (_stricmp(&DropFile[strlen(DropFile)-4], ".BMP") == 0) {
					if (LoadTraceImg(hWnd, DropFile)) {
						fImgLoaded = TRUE;
						InvalidateRect(hDrawWnd, NULL, TRUE);
					}
				}
			}
			DragFinish((HDROP)wParam);
			break;

		case WM_CREATE :
			DragAcceptFiles(hWnd, TRUE);
			comex.dwSize = sizeof(INITCOMMONCONTROLSEX);
			comex.dwICC = ICC_BAR_CLASSES;
			InitCommonControlsEx(&comex);
			hSBWnd = CreateStatusWindow( WS_CHILD | WS_VISIBLE | CCS_BOTTOM | SBARS_SIZEGRIP,
						NULL,
						hWnd,
						IDR_MENU1);
			SendMessage(hSBWnd, SB_SETPARTS, 4, (LPARAM)sbSizes);

			hDrawWnd = CreateWindowEx(WS_EX_CLIENTEDGE,
							"mkv_draw", "",
							WS_CHILD | WS_VISIBLE | WS_HSCROLL | WS_VSCROLL,
							0, 0, ((LPCREATESTRUCT)lParam)->cx, ((LPCREATESTRUCT)lParam)->cy - 25,
							hWnd,
							(HMENU)1010,
							((LPCREATESTRUCT)lParam)->hInstance,
							NULL);
			if (!hDrawWnd) return -1;
			break;

		case WM_SIZE :
			SendMessage(hSBWnd, WM_SIZE, wParam, lParam);
			GetClientRect(hSBWnd, &crect);
			MoveWindow(hDrawWnd, 0, 0,
						LOWORD(lParam) + 1,
						HIWORD(lParam) + 1 - (crect.bottom + 1 - crect.top),
						TRUE);
			break;

		case WM_CLOSE :
			PutFrame(CurrentFrame, 0);
			if (fFileChanged) {
				res = MessageBox(hWnd, 
						"The changes have not been saved.\nDo you wish to save changes?", "Exit Program",
						MB_YESNOCANCEL | MB_ICONWARNING);
				if (res == IDCANCEL) return 0;
				if ((res == IDYES) && !SaveFrames(hWnd, 0, 0, nFrames-1)) return 0;
			}
			DestroyWindow(hWnd);
			return 0;

		case WM_DESTROY :
			PostQuitMessage(0);
			break;

		default :
			return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}




BOOL InitializeProg(HINSTANCE hInstance, int nCmdShow)
{
	WNDCLASSEX myProg;


	hMainInst = hInstance;

	InitCommonControls();

	/* Main window */
	myProg.cbSize =			sizeof(WNDCLASSEX);
	myProg.style =			CS_HREDRAW | CS_VREDRAW;
	myProg.lpszClassName =	"mkv_main";
	myProg.lpfnWndProc =	MainWndProc;
	myProg.cbClsExtra =		0;
	myProg.cbWndExtra =		0;
	myProg.hInstance =		hInstance;
	myProg.hIcon =			LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
	myProg.hCursor =		LoadCursor(NULL, IDC_ARROW);
	myProg.hbrBackground =	(HBRUSH)COLOR_WINDOW;
	myProg.lpszMenuName =	MAKEINTRESOURCE(IDR_MENU1);
	myProg.hIconSm =		NULL;
	if (!RegisterClassEx(&myProg)) return FALSE;

	/* Vector List window */
	myProg.lpszClassName =	"mkv_vect";
	myProg.lpfnWndProc =	VectWndProc;
	myProg.lpszMenuName =	NULL;
	if (!RegisterClassEx(&myProg)) return FALSE;

	/* Frame List window */
	myProg.lpszClassName =	"mkv_frame";
	myProg.lpfnWndProc =	FrameWndProc;
	myProg.lpszMenuName =	NULL;
	if (!RegisterClassEx(&myProg)) return FALSE;

	/* Drawing window */
	myProg.lpszClassName =	"mkv_draw";
	myProg.lpfnWndProc =	DrawWndProc;
	myProg.hCursor =		LoadCursor(NULL, IDC_CROSS);
	myProg.hbrBackground =	GetStockObject(BLACK_BRUSH);
	if (!RegisterClassEx(&myProg)) return FALSE;

	hMainWnd = CreateWindow("mkv_main", "",
						WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
						CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
						NULL,
						NULL,
						hInstance,
						NULL);
	if (!hMainWnd) return FALSE;

	hAccel = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCELERATOR1));
	if (!hAccel) return FALSE;

	ShowWindow(hMainWnd, nCmdShow);
	UpdateWindow(hMainWnd);
	UpdateCaption(3);

	return TRUE;
}




int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg;

	if (!InitializeProg(hInstance, nCmdShow)) return FALSE;

	while (GetMessage(&msg, NULL, 0, 0)) {
		if (!TranslateAccelerator(hMainWnd, hAccel, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	return msg.wParam;
}

